#ifndef CAMERA_
#define CAMERA_

#include <stdio.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "PositionPath.h"
#include "Geometry.h"

#define CAMERA_STANDARD_ANGLE		45.0
#define CAMERA_STANDARD_RATIO		1.33333
#define CAMERA_STANDARD_NEAR		0.1
#define CAMERA_STANDARD_FAR			10000.0

#define DEG2RAD(a)		(a * 0.01745329251994)
#define RAD2DEG(a)		(a * 57.2957795130823)

/**
 * \brief A simple pinhole camera.
 *
 * This class implements a simple pinhole camera. It has attributes like the aspect ratio between the horizontal
 * and vertical size of the capture window, the size of the viewing angle and a near and far cutoff threshold which
 * define the interval in which the objects are visibible to the camera.
 *
 * The camera class also provides means to apply its position and orientation to the current OpenGL matrix as well
 * as some functionality that can be used for viewing frustum culling.
 *
 * It is important to notice, that the viewing direction of this camera is along the positive x axis (and not
 * the z axis as in most other implementations). This choice has been made to make camera placement in world 
 * coordinates simpler as the rotation angles to be applied get more intuitive.
 */
class Camera
{
public:
	/**
	 * Constructor: Initializes the camera with some standard values.
	 *
	 */
	Camera ();

	/**
	 * Destructor.
	 */
	~Camera ();

	/**
	 * Set the attributes of the camera.
	 *
	 * \param angle				The vertical viewing angle of the camera, measured in degrees.
	 * \param aspect_ratio		The horizontal length divided by the vertical length of the viewing-window.
	 * \param near_cutoff		All points that are closer to the camera than this threshold won't get displayed.
	 * \param far_cutoff		All points that are farther away from the camera than this threshold won't get displayed.
	 */
	void SetAttributes (double angle, double aspect_ratio, double near_cutoff, double far_cutoff);

	/**
	 * Sets up the projection matrix of OpenGL.
	 */
	void ApplyAttributes ();

	/**
	 * Sets up the modelview matrix of OpenGL at any given time.
	 *
	 * \param time				Specifies the time which the position shall be applied at.
	 */
	void ApplyPosition (int time);

	/**
	 * Computes the inverse view matrix at any given time.
	 *
	 * \param time				Specifies the time which the inverse view matrix shall be computed at.
	 */
	void ComputeInverseView (int time);

	/**
	 * Returns a pointer to the inverse view matrix. The returned values are only valid if ComputeInverseView (...)
	 * has been called before.
	 *
	 * \return					Pointer to an array, wich contains a 4x4, column-major matrix
	 */
	float * GetInverseView ();

	/**
	 * The path of the camera over time.
	 */
	PositionPath * position;

	/**
	 * Builds the planes of the viewing frustum in world coordinates. Since the camera may change
	 * its position as well as orientation over time, these planes may vary over time.
	 *
	 * \param time				Specifies the time which the view frustum planes shall be built at.
	 */
	void BuildPlanes (int time);

	/** 
	 * Builds the planes of the viewing frustum in the coordinate system right before the projection.
	 * Since the camera position is just another transformation used to express everything our desired 
	 * world coordinates, these planes are time independent.
	 */
	void BuildPlanes2 ();

	/** class members: the view frustum planes in world coordinates. */
	Plane left, right, top, bottom, front, back;

	/** class members: the view frustum planes in the coordinate system right before projection takes place. */
	Plane planes[6];


	/**
	 * Draws the view frustum of the camera by using the planes defined in world coordinates. This routine does not 
	 * display a meaningful frustum unless BuildPlanes (...) has been called before.
	 */
	void DrawFrustum ();

	/**
	 * Draws the view frustum of the camera by using the planes defined in the coordinate system right before the
	 * projection. Since this frustum is independent of the camera position, it must be transformed back into 
	 * world coordinates before it can be drawn. This routine does not display a meaningful frustum unless 
	 * BuildPlanes2 () has been called before.
	 *
	 * \param time				Specifies the time which the view frustum shall be drawn at.
	 */
	void DrawFrustum2 (int time);

	/**
	 * Moves the camera forward by a certain distance.
	 *
	 * \param amount			Specifies the moving distance. A negative value can be used to move backwards.
	 */
	void MoveForward (double amount);

	/**
	 * Moves the camera to the right by a certain distance.
	 *
	 * \param amount			Specifies the moving distance. A negative value can be used to move to the left.
	 */
	void MoveRight (double amount);

	/**
	 * Moves the camera upwards by a certain distance.
	 *
	 * \param amount			Specifies the moving distance. A negative value can be used to move downwards.
	 */
	void MoveUp (double amount);

private:
	/* members to store the attributes of the camera */

	/** viewing angle in degrees */
	double angle;

	/** aspect_ratio */
	double aspect_ratio;

	/** near cutoff value */
	double near_cutoff;

	/** far cutoff value */
	double far_cutoff;

	/** stores the inverse view matrix */
	float inverse_view[16];
};


#endif