#ifndef GRAPHICAL_OBJECT_
#define GRAPHICAL_OBJECT_

#include <stdio.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "Camera.h"
#include "RawObject.h"

#define C_OUTSIDE					-1
#define C_PARTIAL					0
#define C_INSIDE					1

#define C_NONE						0
#define C_SPHERE					1
#define C_BOUNDING_BOX				2
#define C_MIXED						3

/**
 * \brief Describes an object of a scene.
 *
 * This is the basis of all the objects which are rendered within a scene. A graphical object is very similar
 * to a node in a scene graph, with the difference that it does not provide any information about color, texture
 * and other attributes that describe the appearance of an object. It does, however, provide a position (that
 * may vary over time) which determines the local coordinate system spawned by this object.
 *
 * A graphical object may contain any number of child objects, be associated with a RawObject (which is responsible
 * for the appearance), provides some functionality to apply/unapply its local coordinate system to the current 
 * OpenGL matrix and to allow view frustum culling against a camera.
 */
class GraphicalObject
{

//--------------------------------------------------------------------------------------------------------
//									GENERAL
//--------------------------------------------------------------------------------------------------------
public:
	/**
	 * Constructor: Brings the graphical object into a consistent state and associates an object with it.
	 *
	 * \param raw_object				The object with which the graphcial object shall be associated.
	 */
	GraphicalObject (RawObject * raw_object);

	/**
	 * Destructor.
	 */
	~GraphicalObject ();

	/**
	 * The way a graphical object is being drawn has to be implemented by the different types of objects that
	 * inherit from this class.
	 *
	 * \param time					Time at which the object shall be drawn.
	 * \param camera				Camera which captures the object.
	 * \param monitor				Camera which monitors the whole scene, is equal to NULL if we are not in monitoring 
	 *								mode. 
	 * \param show_bounding_volume	States if the bounding volume used for the view frustum culling shall be drawn.
	 * \param culling_mode			Determines the mode used for the view frustum culling (C_NONE, C_BOUNDING_BOX, C_SPHERE
	 *								or C_MIXED).
	 * \param level_of_detail		States wheather level of detail is enabled or not.				
	 */
	virtual void Draw (int time, Camera * camera, Camera * monitor, bool show_bounding_volume, int culling_mode, bool level_of_detail) = NULL;

	/**
	 * The path of this object over time.
	 */
	PositionPath * position;

	/** 
	 * A utility routine that levels out the difference between the own position and the positions of all the children
	 * of this object. The new position is set to the average of all child positions and the child positions are
	 * updated in such a way that no change can be noticed. Note that this routine will create the expected result
	 * only if the children of this object are not interleaved with themselves or other objects (which means 
	 * that the whole structure below this object has to be a tree).
	 */
	void CalculateRelativePositions ();

	/**
	 * Appies the transformation described by an arbitrary position to the current matrix in OpenGL.
	 *
	 * \param position				Position which shall be applied to the current matrix in OpenGL.
	 * \param rotation				Flag indicating wheather it is necessary to apply rotation or not.
	 */
	void ApplyTransformation (const Position & position, bool rotation);

	/**
	 * Undoes the transformation described by an arbitrary position on the current matrix in OpenGL.
	 *
	 * \param position				Position which shall be undone.
	 * \param rotation				Flag indicating wheather it is necessary to undo the rotation or not.
	 */
	void UndoTransformation (const Position & position, bool rotation);

	/**
	 * Prints information about the object to the console. Has to be implemented by all descendants.
	 */
	virtual void PrintToConsole () = NULL;


protected:
	/**
	 * Structure holding all the children of this graphical object.
	 */
	typedef struct ChildListElem
	{
		ChildListElem * next;		/**< Pointer to the structure holding the next child. */
		GraphicalObject * object;	/**< Pointer to the actual child. */
	}ChildList;

	/** Pointer to the first child of this graphical object. */
	ChildList * first_child;

	/** The RawObject that is associated to this graphical object */
	RawObject * raw_object;

private:
	/**
	 * Hide the default constructor.
	 */
	GraphicalObject ();


//--------------------------------------------------------------------------------------------------------
//									CULLING RELATED
//--------------------------------------------------------------------------------------------------------
public:
	/** The radius of the bounding sphere around this graphical object. Used for the view frustum culling. */
	double bounding_radius;

	/** The bounding box around this graphical object. Used for the view frustum culling. */
	BoundingBox bounding_box;

	/**
	 * Routine that updates the bounding radius of this graphical object according to the RawObject associated with 
	 * this graphical object and the position as well as the movement of the children. The object itself and 
	 * all its children will always be inside the sphere defined by the newly created bounding radius at every
	 * point in time.
	 */
	void UpdateBoundingRadius (bool recurse);

	/**
	 * routine that updates the bounding box of this graphical object according to the RawObject associated with
	 * this graphical object and the position as well as the movement of the children. The object itself and
	 * all its children will always be inside the newly calculated bounding box at every point in time. If the
	 * local coordinate systems of the children contain dynamic rotation, a conservative guess about the size
	 * of the new bounding box is being made. If the local coordinate systems of the chidren contain rotation that
	 * stays constant over time, the exact bounding box is being calculated.
	 *
	 * \param recurse				States wheather this routine shall be applied to its children recursifely or not.
	 */
	void UpdateBoundingBox (bool recurse);

protected:
	/**
	 * Returns the result of the view frustum culling test depending on culling mode. Furthermore, this routine
	 * may visualize the bounding volume.
	 *
	 * \param camera				Camera which shall be used for the view frustum culling test.
	 * \param monitor				Camera which is used to monitor the scene during monitoring mode. 
	 * \param culling_mode			Mode of the view frustum culling (C_NONE, C_BOUNDING_BOX, C_SPHERE or C_MIXED).
	 * \param show_volume			Flag indicating whether the bounding volume shall be visualized or not.
	 */
	int CullObject (Camera * camera, Camera * monitor, int culling_mode, bool show_volume);

	/**
	 * Routine to compute the bounding sphere of the graphical object. Has to be implemented by the
	 * descendants of this class.
	 */
	virtual void ComputeBoundingRadius () = NULL;

	/**
	 * Routine to compute the bounding box of the graphical object. Has to be implemented by the descendants
	 * of this class.
	 */
	virtual void ComputeBoundingBox () = NULL;

private:
	/** Corners of the bounding box. Only valid after 'UpdateBoxCorners ()' has been called. */
	BoxCorners	box_corners;

	/** Index of the plane where the object was rejected as being outside at the last time. */
	int last_outside;

	/**
	 * Calculates the corners of the bounding box.
	 */
	void UpdateBoxCorners ();

	/**
	 * Performs a view frustum culling test with a sphere (uses global coordinates to cull),
	 * applied during monitoring mode.
	 *
	 * \param camera				The camera capturing the scene.
	 * \param monitor				The camera monitoring the scene.
	 */
	int SphereCulling (Camera * camera, Camera * monitor);

	/**
	 * Performs a view frustum culling test with a sphere (uses coordinate system right before the projection takes
	 * place), applied when not in monitoring mode.
	 *
	 * \param camera				The camera capturing the scene.
	 */
	int SphereCulling2 (Camera * camera);

	/**
	 * Performs a view frustum culling test with the bounding box of this object (uses global coordinates to cull),
	 * applied during monitoring mode.
	 *
	 * \param camera				The camera capturing the scene.
	 * \param monitor				The camera monitoring the scene.
	 */
	int BoxCulling (Camera * camera, Camera * monitor);

	/**
	 * Performs a view frustum culling test with the bounding box of this object (uses coordinate system right 
	 * before the projection takes place), applied when not in monitoring mode.
	 *
	 * \param camera				The camera capturing the scene.
	 */
	int BoxCulling2 (Camera * camera);

	/**
	 * Draws the bounding sphere around the graphcical object.
	 *
	 * \param culling_result		The result of the view frustum culling test determines the color used for drawing
	 *								the sphere.
	 */
	void DrawBoundingSphere (int culling_result);

	/**
	 * Draws the bounding box around the graphical object.
	 * 
	 * \param culling_result		The result of the view frustum culling test determines the color used for drawing
	 *								the bounding box.
	 */
	void DrawBoundingBox (int culling_result);
};

#endif