#ifndef PERLIN_NOISE_
#define PERLIN_NOISE_

#include <math.h>
#include <stdio.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "Geometry.h"

#define LINEAR_INTERPOLATION		0
#define COSINE_INTERPOLATION		1
#define CUBIC_INTERPOLATION			2


/**
 * \brief Class used to generate 2D perlin noise.
 *
 * PerlinNoise is a class that can be used to generate 2 dimensional perlin noise. It contains some modified,
 * rewritten code from the two internet sources listed below.
 *
 * The whole class is stack based, which means that additional elements can be pushed onto the stack and
 * any operation like generating the noise, transforming the noise or setting parameters is only applied to the
 * top of the stack. Furthermore, there are some combination routines, which combine the top two elements of the
 * stack into one single element and thereby reduce the stack size by one.
 *
 * Every stack element contains two buffers, a buffer with the actual noise and a color buffer. The actual
 * noise buffer can be filled by generating noise (with some parameters). This buffer can then be modified
 * by applying certain transformation routines to each value of the buffer. This buffer can either be exported
 * directly or be written into the color buffer of the stack element. This procedure will generate color values
 * from the noise values by interpolating between the foreground and background color attributes. The color
 * buffers of the stack elements can then either be exported to an array of color values or combined with each other
 * by applying a certain combination routine to each of the color values.
 *
 * For transforming the noise in the noise buffer or combining two color buffers of the stack, there are some
 * basic routines implemented as static members of this class. But the transformations and combinations are in no way
 * restriced to these routines, any routine can be used for transforming and combining the noise as long as it
 * has the correct argument and return value types.
 *
 * Sources: - http://www.animeimaging.com/asp/PerlinNoise.aspx
 *			- http://freespace.virgin.net/hugo.elias/models/m_perlin.htm
 *
 */
class PerlinNoise
{
public:
	/**
	 * Constructor: Brings the object into a consistent state. The dimensions of the perlin noise has to be specified
	 * when the object is created.
	 *
	 * \param size_x					x dimension of the perlin noise.
	 * \param size_y					y dimension of the perlin noise.
	 */
	PerlinNoise (int size_x, int size_y);

	/**
	 * Destructor.
	 */
	~PerlinNoise ();

	/**
	 * Sets the foreground color that will be used when the noise is transformed into color space.
	 * This routine does only affect the top of the stack.
	 *
	 * \param red						Red part of the foreground color.
	 * \param green						Green part of the foreground color.
	 * \param blue						Blue part of the foreground color.
	 * \param alpha						Alpha value of the foreground color.
	 */
	void SetForegroundColor (float red, float green, float blue, float alpha);

	/**
	 * Sets the background color that will be used when the noise is transformed into color space.
 	 * This routine does only affect the top of the stack.
	 *
	 * \param red						Red part of the background color.
	 * \param green						Green part of the background color.
	 * \param blue						Blue part of the background color.
	 * \param alpha						Alpha value of the background color.
	 */
	void SetBackgroundColor (float red, float green, float blue, float alpha);

	/**
	 * Sets the parameters used for the noise generation.
 	 * This routine does only affect the top of the stack.
	 *
	 * \param frequency					Lowest frequency of the noise. This frequency is independent of 
	 *									the size of the noise.
	 * \param persistence				Persistence defines how much higher frequencies are weighted with
	 *									respect to the lower frequencies.
	 * \param octaves					Number of noise patterns that are added to the final noise.
	 * \param smooth					Flag indicating whether the noise shall be smoothed or not.
	 */
	void SetParameters (float frequency, float persistence, int octaves, bool smooth);

	/**
	 * Sets the mode of interpolation used between the bare random noise values.
 	 * This routine does only affect the top of the stack.
	 *
	 * \param mode						Interpolation mode (LINEAR_INTERPOLATION, COSINE_INTERPOLATION or 
	 *									CUBIC_INTERPOLATION).
	 */
	void SetInterpolationMode (int mode);

	/**
	 * Transforms the generated noise with a function that depends on the x and y coordinated of the noise.
	 * x and y, which are fed to this function, are not the dimensions of the noise. They merely depend 
	 * on the frequency times coordinate values between 0 and 1 (0 at top and left/ 1 at bottom and right).
 	 * This routine does only affect the top of the stack.
	 *
	 * \param processor					Pointer to a function which takes an x and y coordinate as well as
	 *									an input value and returns a transformed value.
	 */
	void TransformXY (float (* processor)(float x, float y, float value));

	/**
	 * Transforms the generated noise with a function that takes two parameters a and b. Each value of the
	 * noise is transformed by calling the function with both parameters on each value of the perlin noise.
 	 * This routine does only affect the top of the stack.
	 *
	 * \param a							First parameter that will be fed to the function specified as the
	 *									third argument of this routine.
	 * \param b							Second parameter that will be fed to the function specified as the
	 *									third argument of this routine.
	 * \param processor					Pointer to a function which takes two parameter as well as an input
	 *									value and returns a transformed value.
	 */
	void TransformAB (float a, float b, float (* processor)(float a, float b, float value));

	/**
	 * Generates a two dimensional perlin noise using the actual parameters. Transformations, direct exports
	 * and flushed to the color buffer do not produce meaningful results unless this routine has been called 
	 * first. This routine does only affect the top of the stack.
	 */
	void GenerateNoise ();

	/**
	 * Exports the values of the color buffer on top of the stack to an array of unsigned bytes in RGB format.
	 * This routine does not produce any meaningful result if 'FlushToColorBuffer ()' has not been called on
	 * the element currently on top of the stack.
	 *
	 * \param data_pointer				pointer to an array of unsigned bytes. This array will be filled with
	 *									color values from 0 to 255 in RGB format and its size has to be of at
	 *									least size_x * size_y * 3 elements.
	 */
	void ExportRGB256 (GLubyte * data_pointer);

	/**
	 * Exports the values of the generated perlin noise on top of the stack to an array of single precision 
	 * floating point values.
	 *
	 * \param data_pointer				pointer to an array of single precision floating point values. This
	 *									array will be filled with the perlin noise values and its size has to
	 *									be of at least size_x * size_y elements.
	 */
	void Export (float * data_pointer);

	/**
	 * Returns the x and y dimension of the perlin noise.
	 *
	 * \param dim_x						Variable passed by reference. After the routine, this value will contain
	 *									the size of the perlin noise in the x dimension.
	 * \param dim_y						Variable passed by reference. After the routine, this value will contain
	 *									the size of the perlin noise in the y dimension.
	 */
	void GetDimensions (int &dim_x, int &dim_y);

	/**
	 * Pushes a new element onto the stack.
	 */
	void Push ();

	/**
	 * Flushes the perlin noise to the color buffer by linearly interpolating the values of the perlin noise
	 * between the background and foreground color. Because the values of the perlin noise may be smaller than 0 or
	 * bigger than 1, there is no guarantee that the values written to the color buffer are between the background
	 * and foreground color. This can, however, be achieved by appying a range transform (0..1) on the perlin noise
	 * before flushing it to the color buffer. This routine does only affect the top of the stack.
	 */
	void FlushToColorBuffer ();

	/**
	 * Combines the values of the color buffers of the 2 topmost elements on the stack by applying a function
	 * to all the values in the color buffer. After the two color buffers have been combined, the top element
	 * of the stack is popped from the stack. The resulting color buffer will then be on top of the stack.
	 *
	 * \param combiner					Pointer to a function that takes to values as input and returns an output
	 *									value. This function will be applied to all the values of the two color
	 *									buffers.
	 */
	void CombineColorBuffers (float (* combiner)(float val_1, float val_2));

	/**
	 * Combines the values of the color buffers of the 2 topmost elements on the stack by applying a function
	 * that takes two additional parameters to all the values in the color buffer. After the two color buffers have
	 * been combined, the top element of the stack is popped from the stack. The resulting color buffer will then
	 * be on top of the stack
	 *
	 * \param a							First parameter that shall be fed to the combiner function.
	 * \param b							Second parameter that shall be fed to the combiner function.
	 * \param combiner					Pointer to the combiner function. This function will be fed with
	 *									both parameters as well as the two values of the color buffers and
	 *									it must return the resulting color value.
	 */
	void CombineColorBuffersAB (float a, float b, float (* combiner)(float a, float b, float val_1, float val_2));

	/**
	 * Smoothes the noise at the border such by linearly interpolating the values that are not farther away from
	 * the border that a certain distance. All values on the border are will be zero afterwards.
	 *
	 * \param border					Width of border in which the values are interpolated linearly.
	 */
	void SmoothBorder (int border);


private:
	/**
	 * Hide the default constructor.
	 */
	PerlinNoise ();

	/**
	 * Pops the top element from the stack. Internally used after combining color buffers.
	 */
	void Pop ();

	/**
	 * Checks wheather there are elements on the stack or not.
	 */
	bool IsEmpty ();

	/**
	 * Retrieves a smoothed noise value at one specific point. This is done by evaluating the noise at
	 * (x-1,y-1), (x,y-1), (x+1,y-1), (x-1,y), (x,y), (x+1,y-1), (x-1,y+1), (x,y+1), (x+1,y+1) and weighting
	 * these values in a meaningful way.
	 *
	 * \param x							X coordinate for which the smoothed noise value shall be retrieved.
	 * \param y							Y coordinate for which the smoothed noise value shall be retrieved.
	 */
	float GetSmoothedValue (float x, float y);

	/**
	 * Retrives noise at one specific point.
	 *
	 * \param x							X coordinate for which the noise value shall be retrieved.
	 * \param y							Y coordinate for which the noise value shall be retrieved.
	 */
	float GetNoiseValue (float x, float y);

	/**
	 * Gets the noise at a discrete location.
	 *
	 * \param x							Discrete x coordinate for which the noise value shall be retrieved.
	 * \param y							Discrete y coordinate for which the noise value shall be retrieved.
	 */
	float GetNoise2D (int x, int y);

	/**
	 * Interpolates between two values using the interpolation specified by the interpolation mode.
	 *
	 * \param value_1					First value for the interpolation.
	 * \param value_2					Second value for the interpolation.
	 * \param fraction					Fraction between 0 and 1 used for the interpolation.
	 *									(0: first value, 1: second value).
	 */
	float Interpolate (float value_1, float value_2, float fraction);

	/**
	 * Interpolate linearly between two values.
	 *
	 * \param value_1					First value for the interpolation.
	 * \param value_2					Second value for the interpolation.
	 * \param fraction					Fraction between 0 and 1 used for the interpolation.
	 */
	float LinearInterpolation (float value_1, float value_2, float fraction);

	/**
	 * Use cosine interpolation between two values.
	 *
	 * \param value_1					First value for the interpolation.
	 * \param value_2					Second value for the interpolation.
	 * \param fraction					Fraction between 0 and 1 used for the interpolation.
	 *									(0: first value, 1: second value).
	 */
	float CosineInterpolation (float value_1, float value_2, float fraction);

	/**
	 * Use cubic interpolation between two values.
	 *
     * \param value_1					First value for the interpolation.
	 * \param value_2					Second value for the interpolation.
	 * \param fraction					Fraction between 0 and 1 used for the interpolation.
	 *									(0: first value, 1: second value).
	 */
	float CubicInterpolation (float value_1, float value_2, float fraction);

	/** Size of the noise along the x dimension. */
	int size_x;

	/** Size of the noise along the y dimention. */
	int size_y;

	/** Structure holding an element of the stack. */
	typedef struct StackElem
	{
		float background[4];				/**< Color values for the background color (RGBA). */
		float foreground[4];				/**< Color values for the foreground color (RGBA). */
		float frequency;					/**< Frequency of the perlin noise. */
		float persistence;					/**< Persistence of the perlin noise. */
		int octaves;						/**< Number of different frequencies used to generate the noise. */
		bool smooth;						/**< Flag indicating wheather smoothing thakes place or not. */
		int interpolation;					/**< Interpolation mode. */

		float * perlin_buffer;				/**< Pointer to the generated perlin noise. */
		float * color_buffer;				/**< Pointer to the color buffer. */

		StackElem * next;					/**< Pointer to the next element on the stack. */
	}PerlinStack;

	PerlinStack * top;						/**< Top element of the stack. */


//-------------------------------------------------------------------------------------------
//								PREDEFINED ROUTINES
//-------------------------------------------------------------------------------------------
public:

	/**
	 * Predefined transformation function that can be used with 'TransformAB (...)'.
	 * Makes sure that the resulting value lies within a certain interval.
	 *
	 * \param a							Lower bound of the interval, all returned values will be at 
	 *									least as big as this parameter.
	 * \param b							Upper bound of the interval, all returned values will be smaller
	 *									or equal to this parameter.
	 * \param perlin_xy					Input value.
	 */
	static float RangeTransform (float a, float b, float perlin_xy);

	/**
	 * Predefined transformation function that can be used with 'TransformAB (...)'.
	 * Transforms each point independently with new_value = old_value * a + b.
	 *
	 * \param a							Input value will be multiplied by this value.
	 * \param b							Offset that will be added to the input value.
	 * \param perlin_xy					Input value.
	 */
	static float LinearTransform (float a, float b, float perlin_xy);

	/**
	 * Predefined transformation function that can be used with 'TransformXY (...)'.
	 * Makes the noise appear marbly.
	 *
	 * \param x							X coordinate at which the input value shall be transformed.
	 * \param y							Y coordinate at which the input value shall be transformed.
	 * \param perlin_xy					Input value.
	 */
	static float MarblyTransform (float x, float y, float perlin_xy);

	/**
	 * Predefined transformation function that can be used with 'TransformXY (...)'.
	 * Makes the noise appear grainy.
	 *
	 * \param x							X coordinate at which the input value shall be transformed.
	 * \param y							Y coordinate at which the input value shall be transformed.
	 * \param perlin_xy					Input value.
	 */
	static float GrainyTransform (float x, float y, float perlin_xy);

	/**
	 * Predefined combiner function that can be used with 'CombineColorBuffers'.
	 * Returns the sum of the two input values.
	 *
	 * \param val_1						Value from the first buffer.
	 * \param val_2						Value from the second buffer.
	 * \return							The sum of both values.
	 */
	static float SumCombiner (float val_1, float val_2);

	/**
	 * Predefined combiner function that can be used with 'CombineColorBuffers'.
	 * Returns the product of the two input values.
	 *
	 * \param val_1						Value from the first buffer.
	 * \param val_2						Value from the second buffer.
	 * \return							The product of both values.
	 */
	static float ProductCombiner (float val_1, float val_2);

	/**
	 * Predefined combiner function that can be used with 'CombineColorBuffers'.
	 * Returns the maximum of the two input values.
	 *
	 * \param val_1						Value from the first buffer.
	 * \param val_2						Value from the second buffer.
	 * \return							The maximum of both values.
	 */
	static float MaximumCombiner (float val_1, float val_2);

	/**
	 * Predefined combiner function that can be used with 'CombineColorBuffersAB'.
	 * Returns the weighted sum of the two input values.
	 *
	 * \param weight_1					The weight for the value from the first buffer.
	 * \param weight_2					The weight for the value from the second buffer.
	 * \param val_1						Value from the first buffer.
	 * \param val_2						Value from the second buffer.
	 * \return							The weighted sum of both values (weight_1 * val_1 + weight_2 * val_2).
	 */
	static float WeightedSumCombiner (float weight_1, float weight_2, float val_1, float val_2);
};


#endif