#ifndef RAW_OBJECT_
#define RAW_OBJECT_

#include <stdio.h>
#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "Geometry.h"


/**
 * \brief The base of all objects that get displayed on screen. 
 *
 * A RawObject defines the appearance of an object and contains all its geometric data. The object
 * itself no knowledge about its position and orientation, this information is contained within
 * the GraphicalObjects to which RawObjects get tied to.
 */ 
class RawObject
{
public:
	/**
	 * Constructor: Initializes its members to some default values.
	 */
	RawObject ();

	/**
	 * Destructor.
	 */
	~RawObject ();

	/**
	 * Draws the object. This routine has to be implemented by the descendants of this class.
	 */
	virtual void Draw () = NULL;

	/**
	 * Returns the bounding radius of the object. This routine has to be implemented by the descendants of this class.
	 *
	 * \return						The bounding radius of this object.
	 */
	virtual double GetBoundingRadius () = NULL;

	/**
	 * Returns the bounding box of the object. This routine has to be implemented by the descendants of this class.
	 *
	 * \return						The bounding box of this object.
	 */
	virtual BoundingBox GetBoundingBox () = NULL;

	/**
	 * Set the color of the the object. There is, however, no guarantee that an object will be displayed with that 
	 * color, because the descendants have full responsibility over the appearance of an object.
	 *
	 * \param red					Red part of the material color.
	 * \param green					Green part of the material color.
	 * \param blue					Blue part of the material color.
	 * \param alpha					Alpha value of the material color.
	 */
	void SetColor (float red, float green, float blue, float alpha);

	/**
	 * Set the emission color of the object. There is, however, no guarantee that an object will be displayed with that 
	 * color, because the descendants have full responsibility over the appearance of an object.
	 *
	 * \param red					Red part of the emission color.
	 * \param green					Green part of the emission color.
	 * \param blue					Blue part of the emission color.
	 * \param alpha					Alpha part of the emission color.
	 */
	void SetEmission (float red, float green, float blue, float alpha);

protected:
	/** Internally stores the color of the object. */
	float color[4];

	/** Internally stores the color of the light emitted by this object. */
	float emission[4];
};

#endif